@@ -431,7 +431,7 @@ sub _outs {
     my $raw     = shift;
     my @phrases = (@_);
 
-    Template::Declare->buffer->push( private => (defined wantarray and not wantarray) );
+    Template::Declare->buffer->push( private => (defined wantarray and not wantarray), from => "T::D outs" );
 
     foreach my $item ( grep {defined} @phrases ) {
         my $returned = ref($item) eq 'CODE'
@@ -584,7 +584,7 @@ sub smart_tag_wrapper (&) {
     my $coderef = shift;
 
     Template::Declare->buffer->append("\n");
-    Template::Declare->buffer->push;
+    Template::Declare->buffer->push( from => "T::D tag wrapper");
 
     my %attr = %ATTRIBUTES;
     %ATTRIBUTES = ();                              # prevent leakage
@@ -654,7 +654,7 @@ sub _tag {
 
         local $TAG_NEST_DEPTH = $TAG_NEST_DEPTH + 1;
         %ATTRIBUTES = ();
-        Template::Declare->buffer->push( private => 1 );
+        Template::Declare->buffer->push( private => 1, from => "T::D tag $tag" );
         my $last = join '', map { ref($_) && $_->isa('Template::Declare::Tag') ? $_ : _postprocess($_) } $code->();
         Template::Declare->buffer->append("$last") if not Template::Declare->buffer->length and length $last;
     }
@@ -716,13 +716,15 @@ sub show_page {
     my $args = \@_;
 
     if (defined wantarray) {
-        Template::Declare->buffer->push( private => 1 );
+        Template::Declare->buffer->push( private => 1, from => "T::D path $template" );
         _show_template( $template, 0, $args );
         %ELEMENT_ID_CACHE = ();
         return Template::Declare->buffer->pop;
     } else {
+        Template::Declare->buffer->push( from => "T::D path $template" );
         _show_template( $template, 0, $args );
         %ELEMENT_ID_CACHE = ();
+        Template::Declare->buffer->pop;
         return undef;
     }
 }
@@ -909,7 +911,7 @@ sub stringify {
     my $self = shift;
 
     if ( defined wantarray ) {
-        Template::Declare->buffer->push( private => 1 );
+        Template::Declare->buffer->push( private => 1, from => "T::D stringify" );
         my $returned = $self->();
         return Template::Declare->buffer->pop . $returned;
     } else {
